@@ -1,8 +1,9 @@
+use log::{debug, log_enabled};
 use std::cmp;
 use std::io::{self, Read, Write};
 
 use crate::blocksplitter::{blocksplit, blocksplit_lz77};
-use crate::iter::IsFinalIterator;
+use crate::iter::ToFlagLastIterator;
 use crate::katajainen::length_limited_code_lengths;
 use crate::lz77::{LitLen, Lz77Store, ZopfliBlockState};
 use crate::squeeze::{lz77_optimal, lz77_optimal_fixed};
@@ -12,7 +13,9 @@ use crate::symbols::{
     get_length_symbol_extra_bits,
 };
 use crate::tree::lengths_to_symbols;
-use crate::util::{ZOPFLI_MASTER_BLOCK_SIZE, ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_SIZE};
+use crate::util::{
+    read_to_fill, ZOPFLI_MASTER_BLOCK_SIZE, ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_SIZE,
+};
 use crate::Options;
 
 /// Compresses according to the deflate specification and append the compressed
@@ -26,90 +29,132 @@ use crate::Options;
 /// `in_data`: the input bytes
 /// `out`: pointer to the dynamic output array to which the result is appended. Must
 ///   be freed after use.
-pub fn deflate<R, W>(
-    options: &Options,
-    btype: BlockType,
-    mut in_data: R,
-    insize: u64,
-    out: W,
-) -> io::Result<()>
+pub fn deflate<R, W>(options: &Options, btype: BlockType, mut in_data: R, out: W) -> io::Result<()>
 where
     R: Read,
     W: Write,
 {
-    let mut block_and_window = [0; ZOPFLI_WINDOW_SIZE + ZOPFLI_MASTER_BLOCK_SIZE];
-    let block_and_window_len = block_and_window.len();
+    // The streaming compression algorithm implemented in this function is based on a sliding buffer
+    // that is used and structured as shown in the following ASCII art:
+    //
+    //             0 ┌────────┬───────┬─ ─ ─ ─ ─ ┐
+    // Iteration 0   │ Window │ Block │ Sentinel │
+    //               │  data  │ data  │   byte   │
+    //               └────────┴───────┴─ ─ ─ ─ ─ ┘ WINDOW_SIZE + MASTER_BLOCK_SIZE + 1
+    //                             \            │
+    //      Copy last WINDOW_SIZE   \           │
+    //           block bytes         \          │ Copy sentinel byte to beginning of block data
+    //                                V         V
+    //                               ┌────────┬───────┬─ ─ ─ ─ ─ ┐
+    // Iteration 1                   │ Window │ Block │ Sentinel │
+    //                               │  data  │ data  │   byte   │
+    //                               └────────┴───────┴─ ─ ─ ─ ─ ┘
+    //
+    // The input is read in blocks of MASTER_BLOCK_SIZE + 1 bytes. The last byte is used as
+    // a sentinel to know whether the stream ends at this master block without buffering
+    // previous master blocks. On the beginning of each iteration, the previous sentinel
+    // byte is copied to the beginning of the block data, to consume it for compression.
+    // If we read a full master block plus a new sentinel byte, there is a next master block
+    // at least one byte long, so the just read master block is non-last. If we can't read a
+    // full master block plus a new sentinel byte due to input EOF, the stream ends at this
+    // partial or full master block, so we know it is the last one.
+    //
+    // In addition to the above, a sliding window of the last WINDOW_SIZE bytes of the
+    // previous master block is kept, to allow the DEFLATE stream to reference data from
+    // the former master block, which is key to achieve the expected compression rates.
+
+    const ZOPFLI_WINDOW_AND_BLOCK_SIZE: usize = ZOPFLI_WINDOW_SIZE + ZOPFLI_MASTER_BLOCK_SIZE;
+
+    let mut buffer = vec![0; ZOPFLI_WINDOW_AND_BLOCK_SIZE + 1];
+
+    let mut have_window = false;
+    let mut sentinel_byte = None;
 
     let mut bitwise_writer = BitwiseWriter::new(out);
-    let mut total_bytes_read = 0;
-    let mut bytes_read;
-    let mut i = 0;
-    let mut j = ZOPFLI_MASTER_BLOCK_SIZE;
 
     loop {
-        bytes_read = in_data.read(&mut block_and_window[i..j])?;
-
-        if bytes_read == 0 {
-            // Reached EOF while trying to read a full block.
-            // This means that insize is not a multiple of ZOPFLI_MASTER_BLOCK_SIZE, and the
-            // last block should be smaller.
-            break;
-        }
-
-        total_bytes_read += bytes_read as u64;
-
-        if bytes_read == j - i {
-            // Just read a full block.
-            let final_block = total_bytes_read == insize;
-            let instart = if j == ZOPFLI_MASTER_BLOCK_SIZE {
-                0
-            } else {
-                ZOPFLI_WINDOW_SIZE
-            };
-            deflate_part(
-                options,
-                btype,
-                final_block,
-                &mut block_and_window,
-                instart,
-                j,
-                &mut bitwise_writer,
-            )?;
-
-            if final_block {
-                // No point in trying to read more data.
-                break;
-            } else {
-                // Update buffer start and end indexes accordingly.
-                // Keep a sliding window at the beginning with the latest ZOPFLI_WINDOW_SIZE bytes.
-                // There is no purpose in keeping more than that because only the strings in the
-                // previous ZOPFLI_WINDOW_SIZE bytes can be referenced in a DEFLATE stream.
-                block_and_window.copy_within(j - ZOPFLI_WINDOW_SIZE..j, 0);
-                i = ZOPFLI_WINDOW_SIZE;
-                j = block_and_window_len;
-            }
+        // Inject previous sentinel byte as block data we've just read in this iteration.
+        let sentinel_byte_count = if let Some(sentinel_byte) = sentinel_byte.take() {
+            buffer[ZOPFLI_WINDOW_SIZE] = sentinel_byte;
+            1
         } else {
-            // Continue trying to read a full block.
-            i += bytes_read;
-        }
-    }
-
-    if bytes_read == 0 {
-        let instart = if j == ZOPFLI_MASTER_BLOCK_SIZE {
             0
-        } else {
-            ZOPFLI_WINDOW_SIZE
         };
-        let inend = instart + (total_bytes_read % ZOPFLI_MASTER_BLOCK_SIZE as u64) as usize;
-        deflate_part(
-            options,
-            btype,
-            true,
-            &mut block_and_window,
-            instart,
-            inend,
-            &mut bitwise_writer,
-        )?;
+
+        // Read a new master block of data, plus a new sentinel byte.
+        match read_to_fill(
+            &mut in_data,
+            &mut buffer[ZOPFLI_WINDOW_SIZE + sentinel_byte_count..],
+        )? {
+            (true, _) => {
+                // We have read a full master block and a sentinel byte.
+
+                let (in_data, instart, inend) = if have_window {
+                    (
+                        &buffer[..ZOPFLI_WINDOW_AND_BLOCK_SIZE],
+                        ZOPFLI_WINDOW_SIZE,
+                        ZOPFLI_WINDOW_AND_BLOCK_SIZE,
+                    )
+                } else {
+                    (
+                        &buffer[ZOPFLI_WINDOW_SIZE..ZOPFLI_WINDOW_AND_BLOCK_SIZE],
+                        0,
+                        ZOPFLI_MASTER_BLOCK_SIZE,
+                    )
+                };
+
+                // The next block is at least one byte long due to the sentinel, so we know this block is not the last.
+                deflate_part(
+                    options,
+                    btype,
+                    false,
+                    in_data,
+                    instart,
+                    inend,
+                    &mut bitwise_writer,
+                )?;
+
+                // Update the DEFLATE sliding window with the latest bytes we've just read.
+                buffer.copy_within(
+                    ZOPFLI_WINDOW_AND_BLOCK_SIZE - ZOPFLI_WINDOW_SIZE..ZOPFLI_WINDOW_AND_BLOCK_SIZE,
+                    0,
+                );
+                have_window = true;
+
+                sentinel_byte = Some(buffer[buffer.len() - 1]);
+            }
+            (false, bytes_read) => {
+                // Due to EOF, we either have read a partial master block, or a full block without a sentinel byte.
+                // In both cases, no more blocks will follow.
+
+                let read_block_size = sentinel_byte_count + bytes_read;
+                let (in_data, instart, inend) = if have_window {
+                    (
+                        &buffer[..ZOPFLI_WINDOW_SIZE + read_block_size],
+                        ZOPFLI_WINDOW_SIZE,
+                        ZOPFLI_WINDOW_SIZE + read_block_size,
+                    )
+                } else {
+                    (
+                        &buffer[ZOPFLI_WINDOW_SIZE..ZOPFLI_WINDOW_SIZE + read_block_size],
+                        0,
+                        read_block_size,
+                    )
+                };
+
+                deflate_part(
+                    options,
+                    btype,
+                    true,
+                    in_data,
+                    instart,
+                    inend,
+                    &mut bitwise_writer,
+                )?;
+
+                break;
+            }
+        }
     }
 
     bitwise_writer.finish_partial_bits()
@@ -150,7 +195,6 @@ where
 
             lz77_optimal_fixed(&mut s, in_data, instart, inend, &mut store);
             add_lz77_block(
-                options,
                 btype,
                 final_block,
                 in_data,
@@ -172,7 +216,7 @@ where
     }
 }
 
-#[derive(PartialEq, Copy, Clone, Debug)]
+#[derive(PartialEq, Eq, Copy, Clone, Debug)]
 pub enum BlockType {
     Uncompressed,
     Fixed,
@@ -308,7 +352,7 @@ fn calculate_block_symbol_size_small(
 ) -> usize {
     let mut result = 0;
 
-    debug_assert!(lend - 1 < lz77.size());
+    debug_assert!(lend == lstart || lend - 1 < lz77.size());
 
     for &item in &lz77.litlens[lstart..lend] {
         match item {
@@ -726,8 +770,8 @@ where
 /// `expected_data_size`: the uncompressed block size, used for assert, but you can
 ///   set it to `0` to not do the assertion.
 /// `bitwise_writer`: writer responsible for appending bits
+#[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
 fn add_lz77_block<W>(
-    options: &Options,
     btype: BlockType,
     final_block: bool,
     in_data: &[u8],
@@ -763,12 +807,10 @@ where
 
             let detect_tree_size = bitwise_writer.bytes_written();
             add_dynamic_tree(&ll_lengths, &d_lengths, bitwise_writer)?;
-            if options.verbose {
-                println!(
-                    "treesize: {}",
-                    bitwise_writer.bytes_written() - detect_tree_size
-                );
-            }
+            debug!(
+                "treesize: {}",
+                bitwise_writer.bytes_written() - detect_tree_size
+            );
             (ll_lengths, d_lengths)
         }
     };
@@ -792,18 +834,19 @@ where
     /* End symbol. */
     bitwise_writer.add_huffman_bits(ll_symbols[256], ll_lengths[256])?;
 
-    if options.verbose {
+    if log_enabled!(log::Level::Debug) {
         let uncompressed_size = lz77.litlens[lstart..lend]
             .iter()
             .fold(0, |acc, &x| acc + x.size());
         let compressed_size = bitwise_writer.bytes_written() - detect_block_size;
-        println!(
+        debug!(
             "compressed block size: {} ({}k) (unc: {})",
             compressed_size,
             compressed_size / 1024,
             uncompressed_size
         );
     }
+
     Ok(())
 }
 
@@ -910,6 +953,7 @@ fn get_dynamic_lengths(lz77: &Lz77Store, lstart: usize, lend: usize) -> (f64, Ve
 /// Adds all lit/len and dist codes from the lists as huffman symbols. Does not add
 /// end code 256. `expected_data_size` is the uncompressed block size, used for
 /// assert, but you can set it to `0` to not do the assertion.
+#[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
 fn add_lz77_data<W>(
     lz77: &Lz77Store,
     lstart: usize,
@@ -939,7 +983,7 @@ where
                 let litlen = len as usize;
                 let lls = get_length_symbol(litlen) as u32;
                 let ds = get_dist_symbol(dist as i32) as u32;
-                debug_assert!(litlen >= 3 && litlen <= 288);
+                debug_assert!((3..=288).contains(&litlen));
                 debug_assert!(ll_lengths[lls as usize] > 0);
                 debug_assert!(d_lengths[ds as usize] > 0);
                 bitwise_writer
@@ -961,6 +1005,7 @@ where
     Ok(())
 }
 
+#[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
 fn add_lz77_block_auto_type<W>(
     options: &Options,
     final_block: bool,
@@ -1003,7 +1048,6 @@ where
 
     if uncompressedcost < fixedcost && uncompressedcost < dyncost {
         add_lz77_block(
-            options,
             BlockType::Uncompressed,
             final_block,
             in_data,
@@ -1016,7 +1060,6 @@ where
     } else if fixedcost < dyncost {
         if expensivefixed {
             add_lz77_block(
-                options,
                 BlockType::Fixed,
                 final_block,
                 in_data,
@@ -1028,7 +1071,6 @@ where
             )
         } else {
             add_lz77_block(
-                options,
                 BlockType::Fixed,
                 final_block,
                 in_data,
@@ -1041,7 +1083,6 @@ where
         }
     } else {
         add_lz77_block(
-            options,
             BlockType::Dynamic,
             final_block,
             in_data,
@@ -1111,14 +1152,14 @@ where
     let mut lz77 = Lz77Store::new();
 
     /* byte coordinates rather than lz77 index */
-    let mut splitpoints_uncompressed = Vec::with_capacity(options.blocksplittingmax as usize);
+    let mut splitpoints_uncompressed = Vec::with_capacity(options.maximum_block_splits as usize);
 
     blocksplit(
         options,
         in_data,
         instart,
         inend,
-        options.blocksplittingmax as usize,
+        options.maximum_block_splits,
         &mut splitpoints_uncompressed,
     );
     let npoints = splitpoints_uncompressed.len();
@@ -1128,11 +1169,11 @@ where
     for &item in &splitpoints_uncompressed {
         let mut s = ZopfliBlockState::new(options, last, item);
 
-        let store = lz77_optimal(&mut s, in_data, last, item, options.numiterations);
+        let store = lz77_optimal(&mut s, in_data, last, item, options.iteration_count.get());
         totalcost += calculate_block_size_auto_type(&store, 0, store.size());
 
         // ZopfliAppendLZ77Store(&store, &lz77);
-        debug_assert!(store.size() > 0);
+        debug_assert!(instart == inend || store.size() > 0);
         for (&litlens, &pos) in store.litlens.iter().zip(store.pos.iter()) {
             lz77.append_store_item(litlens, pos);
         }
@@ -1144,11 +1185,11 @@ where
 
     let mut s = ZopfliBlockState::new(options, last, inend);
 
-    let store = lz77_optimal(&mut s, in_data, last, inend, options.numiterations);
+    let store = lz77_optimal(&mut s, in_data, last, inend, options.iteration_count.get());
     totalcost += calculate_block_size_auto_type(&store, 0, store.size());
 
     // ZopfliAppendLZ77Store(&store, &lz77);
-    debug_assert!(store.size() > 0);
+    debug_assert!(instart == inend || store.size() > 0);
     for (&litlens, &pos) in store.litlens.iter().zip(store.pos.iter()) {
         lz77.append_store_item(litlens, pos);
     }
@@ -1158,12 +1199,7 @@ where
         let mut splitpoints2 = Vec::with_capacity(splitpoints_uncompressed.len());
         let mut totalcost2 = 0.0;
 
-        blocksplit_lz77(
-            options,
-            &lz77,
-            options.blocksplittingmax as usize,
-            &mut splitpoints2,
-        );
+        blocksplit_lz77(&lz77, options.maximum_block_splits, &mut splitpoints2);
 
         let mut last = 0;
         for &item in &splitpoints2 {
@@ -1201,7 +1237,7 @@ where
 {
     let in_data = &in_data[instart..inend];
 
-    for (chunk, is_final) in in_data.chunks(65535).is_final() {
+    for (chunk, is_final) in in_data.chunks(65535).flag_last() {
         let blocksize = chunk.len();
         let nlen = !blocksize;
 
@@ -1239,7 +1275,7 @@ where
             bit: 0,
             bp: 0,
             len: 0,
-            out: out,
+            out,
         }
     }
 
