@@ -7,14 +7,19 @@
 //! multiple runs are done with updated cost models to converge to a better
 //! solution.
 
-use std::{cmp, f32, f64};
+use log::{debug, trace};
+use std::{cmp, f64};
 
 use crate::cache::Cache;
 use crate::deflate::{calculate_block_size, BlockType};
 use crate::hash::ZopfliHash;
 use crate::lz77::{find_longest_match, LitLen, Lz77Store, ZopfliBlockState};
-use crate::symbols::{get_dist_extra_bits, get_dist_symbol, get_length_extra_bits, get_length_symbol};
-use crate::util::{ZOPFLI_MAX_MATCH, ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_MASK, ZOPFLI_WINDOW_SIZE};
+use crate::symbols::{
+    get_dist_extra_bits, get_dist_symbol, get_length_extra_bits, get_length_symbol,
+};
+use crate::util::{
+    ZOPFLI_MAX_MATCH, ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_MASK, ZOPFLI_WINDOW_SIZE,
+};
 
 const K_INV_LOG2: f64 = f64::consts::LOG2_E; // 1.0 / log(2.0)
 
@@ -133,7 +138,7 @@ impl SymbolStats {
         fn calculate_and_store_entropy(count: &[usize], bitlengths: &mut [f64]) {
             let n = count.len();
 
-            let sum = count.iter().fold(0, |acc, &x| acc + x);
+            let sum = count.iter().sum();
 
             let log2sum = (if sum == 0 { n } else { sum } as f64).ln() * K_INV_LOG2;
 
@@ -331,7 +336,7 @@ where
         leng = longest_match.length;
 
         // Literal.
-        if i + 1 <= inend {
+        if i < inend {
             let new_cost = costmodel(arr[i] as u32, 0) + costs[j] as f64;
             debug_assert!(new_cost >= 0.0);
             if new_cost < costs[j + 1] as f64 {
@@ -402,6 +407,7 @@ fn trace_backwards(size: usize, length_array: &[u16]) -> Vec<u16> {
 /// `store`: place to output the LZ77 data
 /// returns the cost that was, according to the `costmodel`, needed to get to the end.
 ///     This is not the actual cost.
+#[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
 fn lz77_optimal_run<F, C>(
     s: &mut ZopfliBlockState<C>,
     in_data: &[u8],
@@ -441,7 +447,7 @@ pub fn lz77_optimal_fixed<C>(
     s.blockstart = instart;
     s.blockend = inend;
     let mut h = ZopfliHash::new();
-    let mut costs = Vec::with_capacity(inend - instart - 1);
+    let mut costs = Vec::with_capacity(inend - instart);
     lz77_optimal_run(
         s,
         in_data,
@@ -462,7 +468,7 @@ pub fn lz77_optimal<C>(
     in_data: &[u8],
     instart: usize,
     inend: usize,
-    numiterations: i32,
+    numiterations: u8,
 ) -> Lz77Store
 where
     C: Cache,
@@ -491,7 +497,7 @@ where
     the statistics of the previous run. */
     /* Repeat statistics with each time the cost model from the previous stat
     run. */
-    for i in 0..numiterations {
+    for i in 0..numiterations as i32 {
         currentstore.reset();
         lz77_optimal_run(
             s,
@@ -505,14 +511,15 @@ where
         );
         let cost = calculate_block_size(&currentstore, 0, currentstore.size(), BlockType::Dynamic);
 
-        if s.options.verbose_more || (s.options.verbose && cost < bestcost) {
-            println!("Iteration {}: {} bit", i, cost);
-        }
         if cost < bestcost {
             /* Copy to the output store. */
             outputstore = currentstore.clone();
             beststats = stats;
             bestcost = cost;
+
+            debug!("Iteration {}: {} bit", i, cost);
+        } else {
+            trace!("Iteration {}: {} bit", i, cost);
         }
         let laststats = stats;
         stats.clear_freqs();
